from dataclasses import dataclass
from abc import ABC, abstractmethod

from typing import Dict, Any

from df_and_order.common import DF_TYPE
from df_and_order.helpers import (
    build_class_instance,
    get_file_path_from_module_path,
    FileInspector,
    get_module_path_from_type,
)

TRANSFORM_STEP_MODULE_PATH_KEY = "module_path"
TRANSFORM_STEP_PARAMS_KEY = "params"


@dataclass
class DfTransformStepConfig:
    """
    Stores module path of some DfTransformStep
    as well as its init parameters.
    """

    module_path: str
    params: dict

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DfTransformStepConfig):
            return False

        result = self.module_path == other.module_path and self.params == other.params

        return result

    @staticmethod
    def from_step_type(step_type, params: dict):
        module_path = get_module_path_from_type(py_type=step_type)
        step_config = DfTransformStepConfig(module_path=module_path, params=params)
        return step_config

    @staticmethod
    def from_dict(step_dict: dict):
        module_path = step_dict[TRANSFORM_STEP_MODULE_PATH_KEY]
        params = step_dict.get(TRANSFORM_STEP_PARAMS_KEY) or {}

        step_config = DfTransformStepConfig(module_path=module_path, params=params)
        return step_config

    def to_dict(self) -> dict:
        step_dict: Dict[str, Any] = {
            TRANSFORM_STEP_MODULE_PATH_KEY: self.module_path,
        }

        if len(self.params) > 0:
            step_dict[TRANSFORM_STEP_PARAMS_KEY] = self.params

        return step_dict


class DfTransformStep(ABC):
    """
    Encapsulates logic of some dataframe transformation.
    Every subclass must implement 'transform' method with
    custom logic.
    """

    @staticmethod
    def step_last_modified_ts(step_config: DfTransformStepConfig) -> float:
        file_path = get_file_path_from_module_path(module_path=step_config.module_path)
        result = FileInspector.last_modified_date(file_path=file_path)
        return result

    @staticmethod
    def build_transform(config: DfTransformStepConfig):
        """
        Creates DfTransformStep instance out of config.

        Parameters
        ----------
        config: DfTransformStepConfig
            Contains information about DfTransformStep subclass
            and parameters to init it with.

        Returns
        -------
        DfTransformStep instance.
        """
        params = config.params

        transform = build_class_instance(module_path=config.module_path, init_params=params)
        return transform

    @abstractmethod
    def transform(self, df: DF_TYPE) -> DF_TYPE:
        pass



import pandas as pd
from typing import List

from df_and_order.df_transform_step import DfTransformStep


class TestZeroTransformStep(DfTransformStep):
    def __init__(self, zero_cols: List[str]):
        super().__init__()

        self._zero_cols = zero_cols

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._zero_cols:
            df[col] = 0

        return df



import pandas as pd
from typing import List

from df_and_order.df_transform_step import DfTransformStep, DfTransformStepConfig


class DropColsTransformStep(DfTransformStep):
    """
    Simply drops some undesired columns from a dataframe.
    """

    def __init__(self, cols: List[str]):
        super().__init__()

        self._cols_to_drop = cols

    @staticmethod
    def step_last_modified_ts(step_config: DfTransformStepConfig) -> float:
        return 1589728055.0

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.drop(self._cols_to_drop, axis=1)


class DatesTransformStep(DfTransformStep):
    """
    Converts cols to datetime type
    """

    def __init__(self, cols: List[str]):
        super().__init__()

        self._dates_cols = cols

    @staticmethod
    def step_last_modified_ts(step_config: DfTransformStepConfig) -> float:
        return 1589728055.0

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._dates_cols:
            df[col] = pd.to_datetime(df[col])

        return df



import pytest

from df_and_order.df_transform_step import (
    TRANSFORM_STEP_MODULE_PATH_KEY,
    TRANSFORM_STEP_PARAMS_KEY,
    DfTransformStepConfig,
)
from tests.dates_transform import TestDatesTransformStep


def test_from_step_type():
    step_type = TestDatesTransformStep
    params = {"cols": ["a"]}

    step_config = DfTransformStepConfig.from_step_type(step_type=step_type, params=params)

    assert step_config.module_path == "tests.dates_transform.TestDatesTransformStep"
    assert step_config.params == params


def test_eq():
    module_path = "some/path"
    params = {"param1": 1, "param2": "value"}
    step_dict = {
        TRANSFORM_STEP_MODULE_PATH_KEY: module_path,
        TRANSFORM_STEP_PARAMS_KEY: params,
    }

    step_config = DfTransformStepConfig.from_dict(step_dict=step_dict)
    another_step_config = DfTransformStepConfig.from_dict(step_dict=step_dict.copy())

    assert step_config == another_step_config

    another_step_config.module_path = "bad"

    assert step_config != another_step_config


def test_from_dict():
    module_path = "some/path"
    params = {"param1": 1, "param2": "value"}
    step_dict = {
        TRANSFORM_STEP_MODULE_PATH_KEY: module_path,
        TRANSFORM_STEP_PARAMS_KEY: params,
    }

    step_config = DfTransformStepConfig.from_dict(step_dict=step_dict)

    assert step_config.module_path == module_path
    assert step_config.params == params


def test_to_dict():
    module_path = "some/path"
    params = {"param1": 1, "param2": "value"}
    step_config = DfTransformStepConfig(module_path=module_path, params=params)

    step_dict = step_config.to_dict()

    assert step_dict[TRANSFORM_STEP_MODULE_PATH_KEY] == module_path
    assert step_dict[TRANSFORM_STEP_PARAMS_KEY] == params

    step_config = DfTransformStepConfig(module_path=module_path, params={})
    step_dict = step_config.to_dict()

    assert step_dict.get(TRANSFORM_STEP_PARAMS_KEY) is None



import pytest

from df_and_order.df_transform import (
    DfTransformConfig,
    TRANSFORM_DF_FORMAT_KEY,
    TRANSFORM_IN_MEMORY_KEY,
    TRANSFORM_PERMANENT_KEY,
    TRANSFORM_SOURCE_ID_KEY,
    TRANSFORM_SOURCE_IN_MEMORY_KEY,
)
from df_and_order.df_transform_state import DfTransformState, TRANSFORM_STATE_SOURCE_KEY
from df_and_order.df_transform_step import (
    DfTransformStepConfig,
    TRANSFORM_STEP_MODULE_PATH_KEY,
    TRANSFORM_STEP_PARAMS_KEY,
)


def test_properties(mocker):
    transform = mocker.Mock()
    source_transform = mocker.Mock()

    state = DfTransformState(transform=transform, source_transform=source_transform)

    assert state.transform == transform
    assert state.source_transform == source_transform


@pytest.mark.parametrize("use_source", [True, False], ids=["source", "no_source"])
def test_to_dict(use_source):
    source_transform_id = "source_trans_id"
    transform_id = "trans_id"
    df_format = "df_format"
    in_memory_steps = [
        DfTransformStepConfig(module_path="inmem1", params={"a": 1}),
        DfTransformStepConfig(module_path="inmem2", params={"b": 2}),
    ]
    permanent_steps = [
        DfTransformStepConfig(module_path="perm3", params={}),
        DfTransformStepConfig(module_path="perm4", params={"c": 3}),
    ]
    source_transform = None
    if use_source:
        source_transform = DfTransformConfig(
            transform_id=source_transform_id, df_format=df_format, in_memory_steps=in_memory_steps,
        )

    transform = DfTransformConfig(
        transform_id=transform_id,
        source_id=source_transform_id if use_source else None,
        df_format=df_format,
        permanent_steps=permanent_steps,
    )

    state = DfTransformState(transform=transform, source_transform=source_transform)

    res_transform_id, res_state_dict = state.to_dict()

    test_transform_dict = transform.to_dict()[1]
    test_state_dict = test_transform_dict
    if use_source:
        test_source_dict = source_transform.to_dict()[1]
        test_transform_dict[TRANSFORM_STATE_SOURCE_KEY] = test_source_dict

    assert res_transform_id == transform_id
    assert res_state_dict == test_state_dict


@pytest.mark.parametrize("use_source", [True, False], ids=["source", "no_source"])
def test_from_dict(use_source):
    source_id = "test_source_id" if use_source else None
    df_format = "format"
    in_memory_steps = [
        {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem1", TRANSFORM_STEP_PARAMS_KEY: {"a": 1}},
        {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem2", TRANSFORM_STEP_PARAMS_KEY: {"b": 2}},
    ]
    permanent_steps = [
        {TRANSFORM_STEP_MODULE_PATH_KEY: "perm3",},
        {TRANSFORM_STEP_MODULE_PATH_KEY: "perm4", TRANSFORM_STEP_PARAMS_KEY: {"c": 3}},
    ]
    transform_dict = {
        TRANSFORM_DF_FORMAT_KEY: df_format,
        TRANSFORM_IN_MEMORY_KEY: in_memory_steps,
    }
    state_dict = transform_dict
    source_transform_dict = None
    if use_source:
        source_transform_dict = {
            TRANSFORM_DF_FORMAT_KEY: df_format,
            TRANSFORM_PERMANENT_KEY: permanent_steps,
        }
        state_dict[TRANSFORM_STATE_SOURCE_KEY] = source_transform_dict
        state_dict[TRANSFORM_SOURCE_ID_KEY] = source_id

    transform_id = "trans_id"
    state = DfTransformState.from_dict(transform_id=transform_id, state_dict=state_dict)

    test_transform = DfTransformConfig.from_dict(transform_id=transform_id, transform_dict=transform_dict)
    assert state.transform == test_transform

    if use_source:
        test_source_transform = DfTransformConfig.from_dict(
            transform_id=source_id, transform_dict=source_transform_dict
        )
        assert state.source_transform == test_source_transform



import importlib
import inspect
import os
from typing import Optional, Tuple


def build_class_instance(module_path: str, init_params: Optional[dict] = None):
    """
    Create an object instance from absolute module_path string.

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.
    init_params: optional dict
        These parameters will be used as init parameters for the given type.

    Returns
    -------
    Some object instance
    """
    class_ = get_type_from_module_path(module_path=module_path)
    result = class_(**(init_params or {}))
    return result


def get_type_from_module_path(module_path: str):
    """
    Takes full module path and takes the class
    name to convert it to Python type.

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.

    Returns
    -------
    Python type.
    """
    module_name, class_name = split_module_path(module_path=module_path)

    module = importlib.import_module(module_name)
    class_type = getattr(module, class_name)
    return class_type


def split_module_path(module_path: str) -> Tuple[str, str]:
    """
    Splits module_path strings into a module name and a class name

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.

    Returns
    -------
    Tuple of strings, one for the module name, second for the class name
    """
    components = module_path.split(".")
    class_name = components[-1]
    module_name = ".".join(components[:-1])

    return module_name, class_name


def get_file_path_from_module_path(module_path: str) -> str:
    """
    Returns absolute file path given the module path

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.

    Returns
    -------
    Absolute file path for module path
    """
    module_name, _ = split_module_path(module_path=module_path)
    module = importlib.import_module(module_name)
    file_path = module.__file__
    return file_path


def get_module_path_from_type(py_type) -> str:
    """
    Gets full module path + class name by the given class type

    Parameters
    ----------
    cls: python type
        Class you want to get full module path for

    Returns
    -------
    String representation of full module path like module1.submodule2.ClassName
    """
    result = inspect.getmodule(py_type).__name__ + "." + py_type.__name__  # type: ignore
    return result


class FileInspector:
    @staticmethod
    def last_modified_date(file_path: str) -> float:
        """
        Returns the last modification date unix timestamp for a file at the given path

        Parameters
        ----------
        file_path: str
            Absolute path for any file

        Returns
        -------
        Unix timestamp of the file's last modification date
        """
        result = os.path.getmtime(file_path)
        return result



import pytest
import pandas as pd
import numpy as np
from df_and_order.df_config import DfConfig, DF_ID_KEY, DF_INITIAL_FORMAT_KEY
from df_and_order.df_reader import DfReader
from df_and_order.df_transform import DfTransformConfig, TRANSFORM_IN_MEMORY_KEY
from df_and_order.df_transform_state import TRANSFORM_STATE_SOURCE_KEY
from df_and_order.df_transform_step import DfTransformStepConfig, DfTransformStep
from df_and_order.helpers import FileInspector
from tests.test_df_cache import TestDfCache
from tests.test_df_config import _get_config


@pytest.fixture()
def df_id():
    return "test_id"


@pytest.fixture()
def test_path():
    return "path/to/config/"


@pytest.fixture
def transformed_format():
    return "trans_format"


@pytest.fixture()
def initial_df():
    result = pd.DataFrame(
        {
            "a": ["2020-02-01", "2020-02-02", "2020-02-03"],
            "b": [1, 2, 3],
            "c": ["2020-01-01", "2020-01-02", "2020-01-03"],
        }
    )
    return result


def build_config(mocker, df_id, test_path, initial_format):
    config_dict = {
        DF_ID_KEY: df_id,
        DF_INITIAL_FORMAT_KEY: initial_format,
    }

    result = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)
    return result


@pytest.mark.parametrize("transform_id", ["trans_1", None], ids=["trans", "no_trans"])
@pytest.mark.parametrize("valid_case", [True, False], ids=["valid", "not_valid"])
def test_df_exists(mocker, df_id, test_path, transform_id, valid_case, transformed_format):
    mocker.patch.object(DfConfig, DfConfig.config_exists.__name__, lambda dir_path: True)
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format="init_format")
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)

    if transform_id:
        filename = f"{transform_id}_{df_id}.{transformed_format}"
    else:
        filename = f"{df_id}.{config.initial_df_format}"

    if not valid_case:
        filename += "garbage"

    test_file_path = f"{test_path}{df_id}/{filename}"
    mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: path == test_file_path)

    reader = DfReader(dir_path=test_path, format_to_cache_map={})
    transform = DfTransformConfig(
        transform_id=transform_id,
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.drop_cols_transform.TestDropColsTransformStep", params={"cols_to_drop": ["b"]},
            ),
        ],
    )
    if transform_id:
        # do nothing when it comes to save the config
        mocker.patch.object(DfConfig, DfConfig._save.__name__)
        reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    if valid_case:
        assert reader.df_exists(df_id=df_id, transform_id=transform_id)
    else:
        assert not reader.df_exists(df_id=df_id, transform_id=transform_id)


def test_create_df_config(mocker, df_id, test_path):
    reader = DfReader(dir_path=test_path, format_to_cache_map={})

    initial_format = mocker.Mock()
    metadata = mocker.Mock()
    transform = mocker.Mock()

    config_create_mock = mocker.patch.object(DfConfig, DfConfig.create_config.__name__)

    reader.create_df_config(
        df_id=df_id, initial_df_format=initial_format, metadata=metadata, transform=transform,
    )

    dir_path = _df_dir_path(dir_path=test_path, df_id=df_id)
    config_create_mock.assert_called_with(
        dir_path=dir_path, df_id=df_id, initial_df_format=initial_format, metadata=metadata, transform=transform,
    )


def test_register_transform(mocker, df_id, test_path, transformed_format):
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format="init_format")
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    register_transform_mock = mocker.patch.object(DfConfig, DfConfig.register_transform.__name__)

    transform = mocker.Mock()
    transform.transform_id = "trans_id"
    transform.df_format = transformed_format
    filename = f"{transform.transform_id}_{df_id}.{transformed_format}"

    reader = DfReader(dir_path=test_path, format_to_cache_map={})
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)

    register_transform_mock.assert_called_with(transform=transform, filename=filename)


@pytest.mark.parametrize("existing_transform", [True, False], ids=["existing_transform", "not_existing_transform"])
def test_register_transform_after_read(mocker, df_id, test_path, initial_df, transformed_format, existing_transform):
    initial_df_format = "init_format"
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=initial_df_format)
    mocker.patch.object(DfConfig, DfConfig.config_exists.__name__, lambda dir_path: True)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)

    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        permanent_steps=[
            DfTransformStepConfig(
                module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": ["a"]},
            ),
        ],
    )

    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)

    if existing_transform:
        config.register_transform(transform=transform, filename="test")

    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={initial_df_format: init_cache, transformed_format: transform_cache},
    )
    # ignore states for this test
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    mocker.patch.object(reader, reader._try_to_read_cached_transform.__name__)

    reader.read(df_id=df_id, transform=transform)


def test_read_both_transform_id_and_transform(mocker, df_id, test_path):
    reader = DfReader(dir_path=test_path, format_to_cache_map={})
    with pytest.raises(Exception):
        reader.read(df_id=df_id, transform_id="", transform=mocker.Mock())


def test_read_initial(mocker, df_id, test_path, initial_df):
    init_format = "init_format"
    init_cache = TestDfCache()
    load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    load_mock.return_value = initial_df
    reader = DfReader(dir_path=test_path, format_to_cache_map={init_format: init_cache})
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    df_dir_path = _df_dir_path(dir_path=test_path, df_id=df_id)
    df_path = f"{df_dir_path}/{df_id}.{config.initial_df_format}"

    test_initial_df = reader.read(df_id=df_id)

    load_mock.assert_called_with(path=df_path)

    assert initial_df.equals(test_initial_df)


@pytest.mark.parametrize("permanent", [True, False], ids=["permanent", "not_permanent"])
@pytest.mark.parametrize("from_scratch", [True, False], ids=["from_scratch", "not_from_scratch"])
def test_read_transformed(mocker, df_id, test_path, permanent, from_scratch, initial_df):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    trans_save_mock = mocker.patch.object(transform_cache, transform_cache.save.__name__)
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    transformed_format = "trans_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    # setup transform

    permanent_steps = []
    if permanent:
        permanent_steps = [
            DfTransformStepConfig(
                module_path="tests.drop_cols_transform.TestDropColsTransformStep", params={"cols_to_drop": ["b"]},
            ),
        ]

    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": ["a"]},
            ),
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            ),
        ],
        permanent_steps=permanent_steps,
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    # ignore states for this test
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    mocker.patch.object(reader, reader._try_check_transform.__name__)
    #
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    # override the config getter
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    # just overriding it for the test environment
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)

    df_dir_path = _df_dir_path(dir_path=test_path, df_id=df_id)

    if not from_scratch:
        # if not from scratch we need to stub cached df
        mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)

        transformed_df = initial_df.copy()
        for step_config in transform.permanent_steps:
            step = DfTransformStep.build_transform(config=step_config)
            transformed_df = step.transform(df=transformed_df)

        trans_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
        trans_load_mock.return_value = transformed_df

    transformed_df = reader.read(df_id=df_id, transform=transform)

    # checks

    if from_scratch:
        init_df_path = f"{df_dir_path}/{df_id}.{config.initial_df_format}"
        init_load_mock.assert_called_with(path=init_df_path)

    if permanent:
        test_columns = {"a", "c"}
    else:
        test_columns = {"a", "b", "c"}

    assert set(transformed_df.columns) == test_columns
    assert transformed_df["a"].unique()[0] == 0
    assert transformed_df.select_dtypes(include=[np.datetime64]).columns[0] == "c"

    if from_scratch and permanent:
        df_path = f"{df_dir_path}/{transform.transform_id}_{df_id}.{transformed_format}"
        trans_save_mock.assert_called_with(df=transformed_df, path=df_path)


@pytest.mark.parametrize("is_df_outdated", [True, False], ids=["outdated", "not_outdated"])
@pytest.mark.parametrize("forced", [True, False], ids=["forced", "not_forced"])
def test_read_transformed_check_ts(mocker, df_id, test_path, initial_df, is_df_outdated, forced):
    # if not from scratch we need to stub cached df
    zero_module_path = "tests.zero_transform.TestZeroTransformStep"
    zero_last_modified_ts = 1.0
    drop_module_path = "tests.drop_cols_transform.TestDropColsTransformStep"
    drop_last_modified_ts = 2.0
    if is_df_outdated:
        df_last_modified_ts = min(zero_last_modified_ts, drop_last_modified_ts) - 1.0
    else:
        df_last_modified_ts = max(zero_last_modified_ts, drop_last_modified_ts) + 1.0

    mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)
    mocker.patch.object(
        DfReader, DfReader._df_last_modified_ts.__name__, lambda _, df_id, transform_id: df_last_modified_ts,
    )

    transformed_format = "transformed_format"
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format="init_format")
    # do nothing when it comes to save config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    # override config getter
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)

    permanent_steps = [
        DfTransformStepConfig(module_path=zero_module_path, params={"zero_cols": ["a"]}),
        DfTransformStepConfig(module_path=drop_module_path, params={"cols_to_drop": ["b"]}),
    ]

    transform = DfTransformConfig(
        transform_id="transform_id", df_format=transformed_format, in_memory_steps=[], permanent_steps=permanent_steps,
    )

    trans_cache = TestDfCache()
    trans_load_mock = mocker.patch.object(trans_cache, trans_cache.load.__name__)
    trans_df = mocker.Mock()
    trans_load_mock.return_value = trans_df
    reader = DfReader(dir_path=test_path, format_to_cache_map={transformed_format: trans_cache})
    # ignore states for this test
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    mocker.patch.object(reader, reader._try_check_transform.__name__)
    #
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)

    def last_modified_date(file_path: str):
        if "zero_transform" in file_path:
            return zero_last_modified_ts
        elif "drop_cols_transform" in file_path:
            return drop_last_modified_ts

        raise ValueError("???")

    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, last_modified_date)

    if is_df_outdated and not forced:
        with pytest.raises(Exception):
            reader.read(df_id=df_id, transform=transform, forced=forced)
    else:
        res_trans_df = reader.read(df_id=df_id, transform=transform, forced=forced)
        assert res_trans_df == trans_df


def test_in_memory_shape_when_permanent_is_present(mocker, df_id, test_path, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    # setup transforms
    drop_cols_module_path = "tests.drop_cols_transform.TestDropColsTransformStep"
    dates_module_path = "tests.dates_transform.TestDatesTransformStep"
    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[DfTransformStepConfig(module_path=drop_cols_module_path, params={"cols_to_drop": ["a"]}),],
        permanent_steps=[DfTransformStepConfig(module_path=dates_module_path, params={"dates_cols": ["c"]})],
    )

    # setup config
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    # override the config getter
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)

    with pytest.raises(Exception):
        reader.read(df_id=df_id, transform=transform)


def test_source_id(mocker, df_id, test_path, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    # setup transforms
    drop_cols_module_path = "tests.drop_cols_transform.TestDropColsTransformStep"
    dates_module_path = "tests.dates_transform.TestDatesTransformStep"
    zero_module_path = "tests.zero_transform.TestZeroTransformStep"
    source_transform_id = "source_transform_id"
    source_transform = DfTransformConfig(
        transform_id=source_transform_id,
        df_format=transformed_format,
        in_memory_steps=[DfTransformStepConfig(module_path=dates_module_path, params={"dates_cols": ["c"]})],
    )
    transform = DfTransformConfig(
        transform_id="test_transform_id",
        df_format=transformed_format,
        source_id=source_transform_id,
        source_in_memory_steps=[
            DfTransformStepConfig(module_path=drop_cols_module_path, params={"cols_to_drop": ["a"]})
        ],
        in_memory_steps=[DfTransformStepConfig(module_path=zero_module_path, params={"zero_cols": []})],
    )

    transformed_df = initial_df.copy()
    for step_config in source_transform.in_memory_steps:
        step = DfTransformStep.build_transform(config=step_config)
        transformed_df = step.transform(df=transformed_df)

    trans_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    trans_load_mock.return_value = transformed_df

    # setup config
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    # override the config getter
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    # ignore states for this test
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    mocker.patch.object(reader, reader._try_check_transform.__name__)
    #
    reader.register_transform(df_id=df_id, df_config=config, transform=source_transform)
    transformed_df = reader.read(df_id=df_id, transform=transform)

    assert transformed_df.select_dtypes(include=[np.datetime64]).columns[0] == "c"
    assert "a" not in set(transformed_df.columns)


@pytest.mark.parametrize("valid_state", [True, False], ids=["valid_state", "not_valid_state"])
def test_permanent_transforms_state(mocker, df_id, test_path, valid_state, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    # setup transform

    permanent_steps = [
        DfTransformStepConfig(
            module_path="tests.drop_cols_transform.TestDropColsTransformStep", params={"cols_to_drop": ["b"]},
        ),
    ]

    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": ["a"]},
            ),
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            ),
        ],
        permanent_steps=permanent_steps,
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    mocker.patch.object(reader, DfReader._save_transforms_state_file.__name__)
    transform_dict = transform.to_dict()[1]
    if not valid_state:
        del transform_dict[TRANSFORM_IN_MEMORY_KEY]

    transform_state = {transform.transform_id: transform_dict}
    mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: transform_state)
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)
    mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)

    test_transformed_df = initial_df.copy()
    for step_config in transform.permanent_steps:
        step = DfTransformStep.build_transform(config=step_config)
        test_transformed_df = step.transform(df=test_transformed_df)
    trans_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    trans_load_mock.return_value = test_transformed_df

    if not valid_state:
        with pytest.raises(Exception):
            reader.read(df_id=df_id, transform=transform)
    else:
        transformed_df = reader.read(df_id=df_id, transform=transform)
        assert transformed_df.equals(test_transformed_df)


@pytest.mark.parametrize("valid_state", [True, False], ids=["valid_state", "not_valid_state"])
@pytest.mark.parametrize("forced", [True, False], ids=["forced", "not_forced"])
def test_permanent_source_transforms_state(
    mocker, df_id, test_path, valid_state, initial_df, transformed_format, forced
):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    # setup transform
    source_permanent_steps = [
        DfTransformStepConfig(module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": ["a"]},)
    ]

    permanent_steps = [
        DfTransformStepConfig(
            module_path="tests.drop_cols_transform.TestDropColsTransformStep", params={"cols_to_drop": ["b"]},
        ),
    ]

    source_transform_id = "source_transform_id"
    source_transform = DfTransformConfig(
        transform_id=source_transform_id,
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.zero_transform.TestZeroTransformStep", params={"cols_to_drop": ["b"]},
            ),
        ],
        permanent_steps=source_permanent_steps,
    )

    transform = DfTransformConfig(
        transform_id="transform_id",
        source_id=source_transform_id,
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
        permanent_steps=permanent_steps,
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    mocker.patch.object(reader, DfReader._save_transforms_state_file.__name__)
    transform_dict = transform.to_dict()[1]
    source_transform_dict = source_transform.to_dict()[1]
    if not valid_state:
        del source_transform_dict[TRANSFORM_IN_MEMORY_KEY]

    transform_state = {
        transform.transform_id: transform_dict,
        source_transform.transform_id: source_transform_dict,
    }
    transform_dict[TRANSFORM_STATE_SOURCE_KEY] = source_transform_dict
    mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: transform_state)
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    reader.register_transform(df_id=df_id, df_config=config, transform=source_transform)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)
    mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)

    test_transformed_df = initial_df.copy()
    for step_config in transform.permanent_steps:
        step = DfTransformStep.build_transform(config=step_config)
        test_transformed_df = step.transform(df=test_transformed_df)
    trans_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    trans_load_mock.return_value = test_transformed_df

    if valid_state:
        transformed_df = reader.read(df_id=df_id, transform=transform, forced=forced)
        assert transformed_df.equals(test_transformed_df)
    else:
        if forced:
            transformed_df = reader.read(df_id=df_id, transform=transform, forced=forced)
            assert transformed_df.equals(test_transformed_df)
        else:
            with pytest.raises(Exception):
                reader.read(df_id=df_id, transform=transform, forced=forced)


@pytest.mark.parametrize("df_exists", [True, False], ids=["df_exists", "not_df_exists"])
def test_transforms_state_cleanup(mocker, df_id, test_path, df_exists, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df
    transform_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    transform_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    before_transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
        permanent_steps=[
            DfTransformStepConfig(module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": "a"})
        ],
    )

    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    mocker.patch.object(reader, DfReader._save_transforms_state_file.__name__)
    transform_state = {
        transform.transform_id: before_transform.to_dict()[1],
    }
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: transform_state)
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)

    if df_exists:
        mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)
        mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)
        with pytest.raises(Exception):
            reader.read(df_id=df_id, transform=transform)
    else:
        reader.read(df_id=df_id, transform=transform)
        assert len(transform_state) == 0


def test_transforms_state_save(mocker, df_id, test_path, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df
    transform_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    transform_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    source_transform_id = "source_transform_id"
    source_transform = DfTransformConfig(
        transform_id=source_transform_id,
        df_format=transformed_format,
        permanent_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
    )

    transform = DfTransformConfig(
        transform_id="transform_id",
        source_id=source_transform_id,
        df_format=transformed_format,
        permanent_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    save_mock = mocker.patch.object(reader, DfReader._save_transforms_state_file.__name__)
    transform_dict = transform.to_dict()[1]
    transform_dict[TRANSFORM_STATE_SOURCE_KEY] = source_transform.to_dict()[1]
    test_transform_state = {
        transform.transform_id: transform_dict,
    }
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: {})
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    reader.register_transform(df_id=df_id, df_config=config, transform=source_transform)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)

    reader.read(df_id=df_id, transform=transform)

    save_mock.assert_called_with(df_id=df_id, transforms_state=test_transform_state)


@pytest.mark.parametrize("df_exists", [True, False], ids=["df_exists", "not_df_exists"])
def test_transforms_state_no_file(mocker, df_id, test_path, df_exists, initial_df, transformed_format):
    # setup df cache
    init_cache = TestDfCache()
    transform_cache = TestDfCache()
    init_load_mock = mocker.patch.object(init_cache, init_cache.load.__name__)
    init_load_mock.return_value = initial_df
    transform_load_mock = mocker.patch.object(transform_cache, transform_cache.load.__name__)
    transform_load_mock.return_value = initial_df

    # setup reader
    init_format = "init_format"
    reader = DfReader(
        dir_path=test_path, format_to_cache_map={init_format: init_cache, transformed_format: transform_cache,},
    )

    before_transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
        permanent_steps=[
            DfTransformStepConfig(module_path="tests.zero_transform.TestZeroTransformStep", params={"zero_cols": "a"})
        ],
    )

    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[
            DfTransformStepConfig(
                module_path="tests.dates_transform.TestDatesTransformStep", params={"dates_cols": ["c"]},
            )
        ],
    )

    # setup config

    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format=init_format)
    # do nothing when it comes to save the config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    mocker.patch.object(reader, DfReader._save_transforms_state_file.__name__)
    transform_state = {
        transform.transform_id: before_transform.to_dict()[1],
    }
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: transform_state)
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)
    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, lambda file_path: 1.0)

    if df_exists:
        mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)
        mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)
        with pytest.raises(Exception):
            reader.read(df_id=df_id, transform=transform)
    else:
        reader.read(df_id=df_id, transform=transform)
        assert len(transform_state) == 0


@pytest.mark.parametrize("forced", [True, False], ids=["forced", "not_forced"])
@pytest.mark.parametrize("is_df_outdated", [True, False], ids=["outdated", "not_outdated"])
def test_transform_state_source_check_ts(mocker, df_id, test_path, initial_df, is_df_outdated, forced):
    zero_module_path = "tests.zero_transform.TestZeroTransformStep"
    zero_last_modified_ts = 1.0
    drop_module_path = "tests.drop_cols_transform.TestDropColsTransformStep"
    drop_last_modified_ts = 2.0
    valid_ts = max(zero_last_modified_ts, drop_last_modified_ts) + 1.0
    if is_df_outdated:
        df_last_modified_ts = min(zero_last_modified_ts, drop_last_modified_ts) - 1.0
    else:
        df_last_modified_ts = valid_ts

    transformed_format = "transformed_format"
    config = build_config(mocker=mocker, df_id=df_id, test_path=test_path, initial_format="init_format")
    # do nothing when it comes to save config
    mocker.patch.object(DfConfig, DfConfig._save.__name__)
    # override config getter
    mocker.patch.object(DfReader, DfReader._get_config.__name__, lambda _, df_id: config)

    source_permanent_steps = [
        DfTransformStepConfig(module_path=zero_module_path, params={"zero_cols": ["a"]}),
    ]
    permanent_steps = [
        DfTransformStepConfig(module_path=drop_module_path, params={"cols_to_drop": ["b"]}),
    ]

    source_id = "source_transform_id"
    source_transform = DfTransformConfig(
        transform_id=source_id,
        df_format=transformed_format,
        in_memory_steps=[],
        permanent_steps=source_permanent_steps,
    )

    transform = DfTransformConfig(
        transform_id="transform_id",
        source_id=source_id,
        df_format=transformed_format,
        in_memory_steps=[],
        permanent_steps=permanent_steps,
    )

    mocker.patch.object(DfReader, DfReader.df_exists.__name__, lambda _, df_id, transform_id: True)
    mocker.patch.object(
        DfReader,
        DfReader._df_last_modified_ts.__name__,
        lambda _, df_id, transform_id: df_last_modified_ts if transform_id == source_id else valid_ts,
    )

    trans_cache = TestDfCache()
    trans_load_mock = mocker.patch.object(trans_cache, trans_cache.load.__name__)
    trans_df = mocker.Mock()
    trans_load_mock.return_value = trans_df
    reader = DfReader(dir_path=test_path, format_to_cache_map={transformed_format: trans_cache})
    transform_dict = transform.to_dict()[1]
    source_transform_dict = source_transform.to_dict()[1]
    transform_state = {
        transform.transform_id: transform_dict,
        source_transform.transform_id: source_transform_dict,
    }
    transform_dict[TRANSFORM_STATE_SOURCE_KEY] = source_transform_dict
    mocker.patch.object(DfReader, DfReader._is_file_exists.__name__, lambda path: True)
    mocker.patch.object(reader, DfReader._transforms_state_dicts.__name__, lambda df_id: transform_state)
    # ignore states for this test
    mocker.patch.object(reader, reader._save_transforms_state_file.__name__)
    #
    reader.register_transform(df_id=df_id, df_config=config, transform=source_transform)
    reader.register_transform(df_id=df_id, df_config=config, transform=transform)

    def last_modified_date(file_path: str):
        if "zero_transform" in file_path:
            return zero_last_modified_ts
        elif "drop_cols_transform" in file_path:
            return drop_last_modified_ts

        raise ValueError("???")

    mocker.patch.object(FileInspector, FileInspector.last_modified_date.__name__, last_modified_date)

    if is_df_outdated and not forced:
        with pytest.raises(Exception):
            reader.read(df_id=df_id, transform=transform, forced=forced)
    else:
        res_trans_df = reader.read(df_id=df_id, transform=transform, forced=forced)
        assert res_trans_df == trans_df


def _df_dir_path(dir_path: str, df_id: str) -> str:
    return f"{dir_path}{df_id}"



import pytest

from df_and_order.df_transform import (
    DfTransformConfig,
    TRANSFORM_IN_MEMORY_KEY,
    TRANSFORM_PERMANENT_KEY,
    TRANSFORM_SOURCE_IN_MEMORY_KEY,
    TRANSFORM_SOURCE_ID_KEY,
    TRANSFORM_DF_FORMAT_KEY,
)
from df_and_order.df_transform_step import (
    DfTransformStepConfig,
    TRANSFORM_STEP_MODULE_PATH_KEY,
    TRANSFORM_STEP_PARAMS_KEY,
)


def test_properties(mocker):
    transform_id = "trans_id"
    df_format = "format"
    source_id = "source_id"
    source_in_memory_steps = mocker.Mock()
    in_memory_steps = mocker.Mock()
    permanent_steps = mocker.Mock()

    transform = DfTransformConfig(
        transform_id=transform_id,
        source_id=source_id,
        df_format=df_format,
        source_in_memory_steps=source_in_memory_steps,
        in_memory_steps=in_memory_steps,
        permanent_steps=permanent_steps,
    )

    assert transform.transform_id == transform_id
    assert transform.df_format == df_format
    assert transform.source_id == source_id
    assert transform.source_in_memory_steps == source_in_memory_steps
    assert transform.in_memory_steps == in_memory_steps
    assert transform.permanent_steps == permanent_steps


def test_eq(mocker):
    transform_id = "trans_id"
    df_format = "format"
    source_id = "source_id"
    source_in_memory_steps = mocker.Mock()
    in_memory_steps = mocker.Mock()
    permanent_steps = mocker.Mock()

    transform = DfTransformConfig(
        transform_id=transform_id,
        source_id=source_id,
        df_format=df_format,
        source_in_memory_steps=source_in_memory_steps,
        in_memory_steps=in_memory_steps,
        permanent_steps=permanent_steps,
    )

    another_transform = DfTransformConfig(
        transform_id=transform_id,
        source_id=source_id,
        df_format=df_format,
        source_in_memory_steps=source_in_memory_steps,
        in_memory_steps=in_memory_steps,
        permanent_steps=permanent_steps,
    )

    assert transform == another_transform

    another_transform._transform_id = "bad"

    assert transform != another_transform


@pytest.mark.parametrize("use_source_id", [True, False], ids=["source_id", "no_source_id"])
def test_to_dict(use_source_id):
    transform_id = "trans_id"
    df_format = "df_format"
    source_id = "test_source_id" if use_source_id else None
    source_in_memory_steps = (
        [
            DfTransformStepConfig(module_path="source_inmem1", params={"_a": 1}),
            DfTransformStepConfig(module_path="source_inmem2", params={"_b": 2}),
        ]
        if use_source_id
        else None
    )
    in_memory_steps = [
        DfTransformStepConfig(module_path="inmem1", params={"a": 1}),
        DfTransformStepConfig(module_path="inmem2", params={"b": 2}),
    ]
    permanent_steps = [
        DfTransformStepConfig(module_path="perm3", params={}),
        DfTransformStepConfig(module_path="perm4", params={"c": 3}),
    ]

    transform = DfTransformConfig(
        transform_id=transform_id,
        df_format=df_format,
        source_id=source_id,
        source_in_memory_steps=source_in_memory_steps,
        in_memory_steps=in_memory_steps,
        permanent_steps=permanent_steps,
    )

    res_transform_id, res_transform_dict = transform.to_dict()
    res_source_id = res_transform_dict.get(TRANSFORM_SOURCE_ID_KEY)
    test_transform_dict = {
        TRANSFORM_DF_FORMAT_KEY: df_format,
        TRANSFORM_IN_MEMORY_KEY: [
            {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem1", TRANSFORM_STEP_PARAMS_KEY: {"a": 1},},
            {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem2", TRANSFORM_STEP_PARAMS_KEY: {"b": 2},},
        ],
        TRANSFORM_PERMANENT_KEY: [
            {TRANSFORM_STEP_MODULE_PATH_KEY: "perm3",},
            {TRANSFORM_STEP_MODULE_PATH_KEY: "perm4", TRANSFORM_STEP_PARAMS_KEY: {"c": 3},},
        ],
    }

    if use_source_id:
        test_transform_dict[TRANSFORM_SOURCE_ID_KEY] = source_id
        test_transform_dict[TRANSFORM_SOURCE_IN_MEMORY_KEY] = [
            {TRANSFORM_STEP_MODULE_PATH_KEY: "source_inmem1", TRANSFORM_STEP_PARAMS_KEY: {"_a": 1},},
            {TRANSFORM_STEP_MODULE_PATH_KEY: "source_inmem2", TRANSFORM_STEP_PARAMS_KEY: {"_b": 2},},
        ]

    assert res_transform_id == transform_id
    assert res_source_id == source_id
    assert res_transform_dict == test_transform_dict


@pytest.mark.parametrize("use_source_id", [True, False], ids=["source_id", "no_source_id"])
def test_from_dict(use_source_id):
    source_id = "test_source_id" if use_source_id else None
    df_format = "format"
    source_in_memory_steps = (
        [
            {TRANSFORM_STEP_MODULE_PATH_KEY: "source_inmem1", TRANSFORM_STEP_PARAMS_KEY: {"_a": 1},},
            {TRANSFORM_STEP_MODULE_PATH_KEY: "source_inmem2", TRANSFORM_STEP_PARAMS_KEY: {"_b": 2},},
        ]
        if use_source_id
        else None
    )
    in_memory_steps = [
        {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem1", TRANSFORM_STEP_PARAMS_KEY: {"a": 1}},
        {TRANSFORM_STEP_MODULE_PATH_KEY: "inmem2", TRANSFORM_STEP_PARAMS_KEY: {"b": 2}},
    ]
    permanent_steps = [
        {TRANSFORM_STEP_MODULE_PATH_KEY: "perm3",},
        {TRANSFORM_STEP_MODULE_PATH_KEY: "perm4", TRANSFORM_STEP_PARAMS_KEY: {"c": 3}},
    ]
    transform_dict = {
        TRANSFORM_DF_FORMAT_KEY: df_format,
        TRANSFORM_IN_MEMORY_KEY: in_memory_steps,
        TRANSFORM_PERMANENT_KEY: permanent_steps,
    }

    if use_source_id:
        transform_dict[TRANSFORM_SOURCE_ID_KEY] = source_id
        transform_dict[TRANSFORM_SOURCE_IN_MEMORY_KEY] = source_in_memory_steps

    transform_id = "trans_id"
    transform = DfTransformConfig.from_dict(transform_id=transform_id, transform_dict=transform_dict)

    assert transform.df_format == df_format
    assert transform.transform_id == transform_id
    if use_source_id:
        for original_dict, step in zip(source_in_memory_steps, transform.source_in_memory_steps):
            assert original_dict[TRANSFORM_STEP_MODULE_PATH_KEY] == step.module_path
            assert original_dict[TRANSFORM_STEP_PARAMS_KEY] == step.params

    for original_dict, step in zip(in_memory_steps, transform.in_memory_steps):
        assert original_dict[TRANSFORM_STEP_MODULE_PATH_KEY] == step.module_path
        assert original_dict[TRANSFORM_STEP_PARAMS_KEY] == step.params

    for original_dict, step in zip(permanent_steps, transform.permanent_steps):
        assert original_dict[TRANSFORM_STEP_MODULE_PATH_KEY] == step.module_path
        if original_dict.get(TRANSFORM_STEP_PARAMS_KEY):
            assert original_dict[TRANSFORM_STEP_PARAMS_KEY] == step.params



from dataclasses import dataclass
from typing import Optional, Tuple

from df_and_order.df_transform import DfTransformConfig


TRANSFORM_STATE_SOURCE_KEY = "source_transform"


@dataclass
class DfTransformState:
    transform: DfTransformConfig
    source_transform: Optional[DfTransformConfig]

    def to_dict(self) -> Tuple[str, dict]:
        """
        Converts an instance to serializable dictionary

        Returns
        -------
        tuple of str and dict, transform_id and transform's dictionary representation.
        """
        transform_id, transform_dict = self.transform.to_dict()
        result = transform_dict

        if self.source_transform:
            _, source_dict = self.source_transform.to_dict()
            result[TRANSFORM_STATE_SOURCE_KEY] = source_dict

        return transform_id, result

    @staticmethod
    def from_dict(transform_id: str, state_dict: dict):
        """
        Builds DfTransformState instance out of serialized dictionary.

        Parameters
        ----------
        transform_id: optional str
            Unique identifier of a transform.
        state_dict:
            Dictionary representation of DfTransformState

        Returns
        -------
        DfTransformConfig instance.
        """
        state_dict = state_dict.copy()
        source_dict = state_dict.get(TRANSFORM_STATE_SOURCE_KEY)
        source_transform = None
        if source_dict:
            del state_dict[TRANSFORM_STATE_SOURCE_KEY]

        transform_dict = state_dict
        transform = DfTransformConfig.from_dict(transform_id=transform_id, transform_dict=transform_dict)

        if source_dict:
            source_transform = DfTransformConfig.from_dict(transform_id=transform.source_id, transform_dict=source_dict)
        result = DfTransformState(transform=transform, source_transform=source_transform)
        return result



import os
import yaml
from typing import Optional, Dict, List, Any

from df_and_order.common import DF_TYPE
from df_and_order.df_transform import DfTransformConfig
from df_and_order.df_config import DfConfig
from df_and_order.df_cache import DfCache
from df_and_order.df_transform_state import DfTransformState
from df_and_order.df_transform_step import DfTransformStepConfig, DfTransformStep
from df_and_order.helpers import FileInspector, get_type_from_module_path


class DfReader:
    """
    Reads a dataframe from the disk using a config file.
    Helps you to organize and document your dataframes
    for easy access and reproducibility.

    Parameters
    ----------
    dir_path: str
        Absolute path where your dataframe is supposed to be saved.
    format_to_cache_map: dict
        To be able to work with various formats DfReader must know
        how to read&save them. Provide a map where a format extension is used as a key
        and DfCache instance as a value. See DfCache class documentation for the details.
    """

    def __init__(self, dir_path: str, format_to_cache_map: Dict[str, DfCache]):
        self._dir_path = dir_path
        self._format_to_cache_map = format_to_cache_map

    def df_exists(self, df_id: str, transform_id: Optional[str] = None) -> bool:
        """
        Checks whether a dataframe file exists at the provided path.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            If you want to check whether a transformed version of your dataframe
            is persisted on the disk you may pass its unique identifier that
            matches the one in the config file.

        Returns
        -------
        True if the dataframe exists, False otherwise
        """
        df_dir_path = self._df_dir_path(df_id=df_id)
        if not DfConfig.config_exists(dir_path=df_dir_path):
            return False

        df_config = self._get_config(df_id=df_id)
        df_path = self._df_path(df_config=df_config, df_id=df_id, transform_id=transform_id)
        return DfReader._is_file_exists(path=df_path)

    def _df_last_modified_ts(self, df_id: str, transform_id: Optional[str] = None) -> float:
        df_config = self._get_config(df_id=df_id)
        df_path = self._df_path(df_config=df_config, df_id=df_id, transform_id=transform_id)
        result = FileInspector.last_modified_date(file_path=df_path)
        return result

    @staticmethod
    def _is_file_exists(path: str):
        return os.path.exists(path=path)

    def create_df_config(
        self,
        df_id: str,
        initial_df_format: str,
        metadata: Optional[dict] = None,
        transform: Optional[DfTransformConfig] = None,
    ):
        """
        Just a forwarding to DfConfig method, see docs in DfConfig.
        """
        DfConfig.create_config(
            dir_path=self._df_dir_path(df_id=df_id),
            df_id=df_id,
            initial_df_format=initial_df_format,
            metadata=metadata,
            transform=transform,
        )

    def register_transform(self, df_id: str, df_config: DfConfig, transform: DfTransformConfig):
        """
        Forms a filename for the given dataframe and adds a new transform to the config file if possible.
        In general it's just a forwarding to DfConfig method, see docs in DfConfig.
        """
        filename = self.df_filename(df_config=df_config, df_id=df_id, transform=transform)
        df_config.register_transform(transform=transform, filename=filename)

    def read(
        self,
        df_id: str,
        transform_id: Optional[str] = None,
        transform: Optional[DfTransformConfig] = None,
        forced: bool = False,
    ) -> DF_TYPE:
        """
        Reads a dataframe from the disk. If you want a transformed version of your dataframe,
        but it's still not persisted, it first creates it and then reads it into memory.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            Unique identifier of the desired transform.
        transform: optional DfTransformConfig
            Object that describes all the transformation steps required.
        forced: bool
            If True, no safety checks will be performed,
            so you take the responsibility for data consistency.

        Returns
        -------

        """
        if transform_id and transform:
            raise AttributeError("Provide either transform_id or transform_config")

        df_config = self._get_config(df_id=df_id)
        self._update_transforms_state(df_id=df_id)

        is_transform = transform_id or transform
        if is_transform:
            if not transform:
                assert isinstance(transform_id, str)
                transform = df_config.transforms_by(transform_id=transform_id)

            return self._read_transformed(df_id=df_id, transform=transform, df_config=df_config, forced=forced)

        return self._read_initial(df_id=df_id, df_config=df_config)

    def _read_initial(self, df_id: str, df_config: DfConfig) -> DF_TYPE:
        """
        Reads the original dataframe from the disk
        """
        df_format = df_config.initial_df_format
        return self._read_df(df_id=df_id, df_format=df_format, df_config=df_config)

    def _read_transformed(
        self, df_id: str, transform: DfTransformConfig, df_config: DfConfig, forced: bool = False,
    ) -> DF_TYPE:
        """
        Reads the transformed dataframe from the disk or creates it if needed.
        """
        transform_id = transform.transform_id

        self.register_transform(df_id=df_id, df_config=df_config, transform=transform)

        transformed_df_exists = self.df_exists(df_id=df_id, transform_id=transform_id)
        if transformed_df_exists:
            return self._try_to_read_cached_transform(
                df_id=df_id, df_config=df_config, forced=forced, transform=transform
            )

        df = self._read_source_df_for_transform(df_id=df_id, df_config=df_config, forced=forced, transform=transform)
        df = self._apply_df_transforms(df_id=df_id, df=df, df_config=df_config, transform=transform)

        return df

    def _read_source_df_for_transform(
        self, df_id: str, df_config: DfConfig, forced: bool, transform: DfTransformConfig,
    ):
        if transform.source_id:
            source_transform = df_config.transforms_by(transform_id=transform.source_id)
            df = self.read(df_id=df_id, transform=source_transform, forced=forced)
        else:
            df = self._read_initial(df_id=df_id, df_config=df_config)
        return df

    def _try_to_read_cached_transform(
        self, df_id: str, df_config: DfConfig, forced: bool, transform: DfTransformConfig,
    ):
        if transform.source_id:
            source_transform = df_config.transforms_by(transform_id=transform.source_id)
            self._try_check_transform(
                df_id=df_id, transform=source_transform, forced=forced, child_transform=transform,
            )

        self._try_check_transform(df_id=df_id, transform=transform, forced=forced)
        return self._read_cached_transformed(df_id=df_id, transform=transform, df_config=df_config, forced=forced)

    def _apply_df_transforms(self, df_id: str, df: DF_TYPE, df_config: DfConfig, transform: DfTransformConfig):
        transform_id = transform.transform_id
        df_format = transform.df_format
        both_transform_types_are_present = transform.in_memory_steps and transform.permanent_steps
        df_shape_before = df.shape
        if transform.source_in_memory_steps:
            df = DfReader._apply_transform_steps(df=df, steps=transform.source_in_memory_steps)
        if transform.in_memory_steps:
            df = DfReader._apply_transform_steps(df=df, steps=transform.in_memory_steps)
            df_shape_after_in_memory = df.shape
            df_shape_has_changed = df_shape_before != df_shape_after_in_memory
            if both_transform_types_are_present and df_shape_has_changed:
                raise Exception(
                    "Error: A permanent transform is also present, hence your "
                    "in-memory transform can't modify the shape of the initial df."
                )
        if transform.permanent_steps:
            df = DfReader._apply_transform_steps(df=df, steps=transform.permanent_steps)

            df_path = self._df_path(df_config=df_config, df_id=df_id, transform_id=transform_id)
            df_cache = self._get_df_cache(df_format=df_format)
            df_cache.save(df=df, path=df_path)
            source_transform = None
            if transform.source_id:
                source_transform = df_config.transforms_by(transform_id=transform.source_id)
            state = DfTransformState(transform=transform, source_transform=source_transform)
            self._save_transform_state(df_id=df_id, state=state)

        return df

    def _save_transform_state(self, df_id: str, state: DfTransformState):
        transforms_state = self._transforms_state_dicts(df_id=df_id)
        transform_id = state.transform.transform_id
        assert transforms_state.get(transform_id) is None
        transforms_state[transform_id] = state.to_dict()[1]
        self._save_transforms_state_file(df_id=df_id, transforms_state=transforms_state)

    def _transforms_states(self, df_id: str) -> Dict[str, DfTransformState]:
        state_dicts = self._transforms_state_dicts(df_id=df_id)
        result = {}

        for transform_id, state_dict in state_dicts.items():
            state = DfTransformState.from_dict(transform_id=transform_id, state_dict=state_dict)
            result[transform_id] = state

        return result

    def _transforms_state_dicts(self, df_id: str) -> Dict[str, Any]:
        file_path = self._transforms_state_file_path(df_id=df_id)
        result_dict = {}
        if os.path.exists(file_path):
            with open(file_path) as config_file:
                result_dict = yaml.safe_load(config_file)

        return result_dict

    def _save_transforms_state_file(self, df_id: str, transforms_state: Dict[str, Any]):
        with open(self._transforms_state_file_path(df_id=df_id), "w") as config_file:
            yaml.dump(transforms_state, config_file)

    def _transforms_state_file_path(self, df_id: str) -> str:
        path = self._df_dir_path(df_id=df_id)
        file_path = os.path.join(path, ".transforms_state.yaml")
        return file_path

    # TODO: all the checks should be extracted from the reader
    def _try_check_transform(
        self,
        df_id: str,
        transform: DfTransformConfig,
        forced: bool,
        child_transform: Optional[DfTransformConfig] = None,
    ):
        if child_transform and not self._can_start_transform_state_check(df_id=df_id, transform=transform):
            return

        if child_transform:
            # in case of child transform we better check the source separately
            # e.g. a source config can match source_transform in a child's state
            # but the source file was generated using different config before
            self._try_check_transform(df_id=df_id, transform=transform, forced=forced)

        self._check_transform(df_id=df_id, transform=transform, forced=forced, child_transform=child_transform)

    def _check_transform(
        self,
        df_id: str,
        transform: DfTransformConfig,
        forced: bool,
        child_transform: Optional[DfTransformConfig] = None,
    ):
        """
        Performs various checks for the given transform
        """
        self._check_ts(df_id=df_id, forced=forced, transform=transform)

        # which transform are we checking
        transform_id = self._transform_state_id(transform=transform, child_transform=child_transform)

        # saved state for the transform
        transform_state = self._transform_state(df_id=df_id, transform_id=transform_id)

        # is it possible to check such a state
        can_proceed = self._can_proceed_with_transform_state(
            transform_id=transform.transform_id, transform_state=transform_state, forced=forced
        )

        if not can_proceed:
            # user forced the check, so exiting early
            return

        # now we are sure that transform_state is present, for mypy
        assert transform_state

        # ready to truly check the state
        self._check_transform_state(
            transform_state=transform_state, transform=transform, forced=forced, child_transform=child_transform
        )

    def _can_start_transform_state_check(self, df_id: str, transform: DfTransformConfig) -> bool:
        if len(transform.permanent_steps) == 0:
            return False

        if not self.df_exists(df_id=df_id, transform_id=transform.transform_id):
            # at first I wanted to throw an exception here, but if the parent transform
            # is no longer present on the disk, such fact shouldn't prevent you from using a
            # serialized child transform. It would be a problem if persisted parent transform
            # is incompatible with the one used to generate a child transform's file
            # P.S. at least I understand it...
            return False

        return True

    @staticmethod
    def _transform_state_id(transform: DfTransformConfig, child_transform: Optional[DfTransformConfig]):
        if child_transform:
            return child_transform.transform_id
        return transform.transform_id

    def _transform_state(self, df_id: str, transform_id: str) -> Optional[DfTransformState]:
        transforms_state = self._transforms_states(df_id=df_id)
        transform_state = transforms_state.get(transform_id)

        return transform_state

    @staticmethod
    def _can_proceed_with_transform_state(
        transform_id: str, transform_state: Optional[DfTransformState], forced: bool
    ) -> bool:
        """
        Checks whether transform_state exists on disk. If not, we can't perform checks.
        If forced, only warning will be shown, otherwise it raises Exception.
        """
        if not transform_state:
            base_warning = f"{transform_id} transform was " f"persisted with an unknown configuration"
            if forced:
                print(
                    f"VERY IMPORTANT WARNING: {base_warning}, but "
                    f"reading it anyway because the operation was forced."
                )
                return False

            raise Exception(f"{base_warning}, can't safely load it")

        return True

    def _check_transform_state(
        self,
        transform_state: DfTransformState,
        transform: DfTransformConfig,
        forced: bool,
        child_transform: Optional[DfTransformConfig] = None,
    ):
        # if child_transform is present - we are interested in its source_transform
        # saved state since then the transform parameter represents source's current state
        if child_transform:
            transform_from_state = transform_state.source_transform
        else:
            transform_from_state = transform_state.transform

        if transform_from_state:
            # the only thing left is to compare the saved state with the current one
            # P.S. It would be nicer if I checked NOT case and raised an exception there,
            # but I need to convince mypy that transform_from_state is present
            self._check_transform_with_state(
                transform=transform, transform_from_state=transform_from_state, forced=forced
            )
            return

        raise Exception(f"Critical: couldn't get the transform from transform_state: {transform.transform_id}")

    @staticmethod
    def _check_transform_with_state(
        transform: DfTransformConfig, transform_from_state: DfTransformConfig, forced: bool
    ):
        valid = transform_from_state.to_dict()[1] == transform.to_dict()[1]

        if not valid:
            base_warning = (
                f"You've changed the df_config.yaml file of {transform.transform_id}"
                f" transform, so it's incompatible with the persisted version"
            )
            if forced:
                print(
                    f"VERY IMPORTANT WARNING: {base_warning}, but "
                    f"reading it anyway because the operation was forced."
                )
                return

            raise Exception(base_warning)

    def _update_transforms_state(self, df_id: str):
        """
        When a file of a persisted transform gets deleted -
        the method removes the transform's config from saved state config
        """
        transforms_state = self._transforms_state_dicts(df_id=df_id)

        transform_ids_to_delete = []
        for transform_id, transform_dict in transforms_state.items():
            transform = DfTransformConfig.from_dict(transform_id=transform_id, transform_dict=transform_dict)
            df_config = self._get_config(df_id=df_id)
            filename = DfReader.df_filename(df_config=df_config, df_id=df_id, transform=transform)
            file_path = self._df_dir_path(df_id=df_id, filename=filename)
            if not DfReader._is_file_exists(path=file_path):
                transform_ids_to_delete.append(transform_id)

        for transform_id in transform_ids_to_delete:
            del transforms_state[transform_id]

        self._save_transforms_state_file(df_id=df_id, transforms_state=transforms_state)

    def _read_cached_transformed(
        self, df_id: str, transform: DfTransformConfig, df_config: DfConfig, forced: bool = False,
    ) -> DF_TYPE:
        df_format = transform.df_format
        transform_id = transform.transform_id

        if transform.permanent_steps:
            self._check_ts(df_id=df_id, forced=forced, transform=transform)

        df = self._read_df(df_config=df_config, df_id=df_id, df_format=df_format, transform_id=transform_id,)

        if transform.in_memory_steps:
            df = DfReader._apply_transform_steps(df=df, steps=transform.in_memory_steps)

        return df

    def _check_ts(self, df_id: str, forced: bool, transform: DfTransformConfig):
        """
        If a python file with a transform was changed after a transform
        was persisted on disk - we can no longer guarantee that the persisted file is still relevant.
        This method checks timestamp and raises Exception if needed.
        ( modifications in the file could bring some drastic changes in the transform itself ).
        """
        if len(transform.permanent_steps) == 0:
            return

        # if the code of one of the steps was modified since the transformed dataframe
        # was cached - we need to stop and warn a user about the need of regenerating it
        df_last_modified_date = self._df_last_modified_ts(df_id=df_id, transform_id=transform.transform_id)

        def _filter_func(step_config: DfTransformStepConfig):
            # built-in types override the method to provide true last modification date
            step_class = get_type_from_module_path(module_path=step_config.module_path)
            step_last_modified_ts = step_class.step_last_modified_ts(step_config=step_config)
            result = step_last_modified_ts > df_last_modified_date
            return result

        outdated_steps = list(filter(_filter_func, transform.permanent_steps))
        if len(outdated_steps) > 0:
            steps_module_paths = [step.module_path for step in outdated_steps]
            base_warning = (
                f"{steps_module_paths} steps of {transform.transform_id} "
                f"transform were changed since the df was generated"
            )
            if forced:
                print(f"Warning: {base_warning}, reading it anyway because the operation was forced")
            else:
                raise Exception(f"{base_warning}, " "delete the file and try again to regenerate the df.")

    @staticmethod
    def _apply_transform_steps(df: DF_TYPE, steps: List[DfTransformStepConfig]) -> DF_TYPE:
        """
        Applies all the steps for a transformation on the given dataframe.

        Parameters
        ----------
        df: DF_TYPE
            Initial dataframe to perform transformations on.
        steps: list of DfTransformStepConfig
            List of objects that represent a step of the whole transformation.

        Returns
        -------
        DF_TYPE, fully transformed initial dataframe
        """
        transform_steps = [DfTransformStep.build_transform(config=step_config) for step_config in steps]
        for transform in transform_steps:
            df = transform.transform(df=df)

        return df

    def _read_df(self, df_id: str, df_format: str, df_config: DfConfig, transform_id: Optional[str] = None,) -> DF_TYPE:
        """
        General method for reading a dataframe from the disk.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        df_format: str
            Format in which dataframe was saved.
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        transform_id: optional str
            Unique identifier of the desired transform.

        Returns
        -------
        DF_TYPE, the requested dataframe
        """
        df_path = self._df_path(df_config=df_config, df_id=df_id, transform_id=transform_id)

        df_cache = self._get_df_cache(df_format=df_format)
        df = df_cache.load(path=df_path)

        return df

    def _df_path(self, df_config: DfConfig, df_id: str, transform_id: Optional[str] = None) -> str:
        """
        Forms a path to the dataframe.

        Parameters
        ----------
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            Unique identifier of the desired transform.

        Returns
        -------
        str, absolute path to the dataframe
        """
        transform = None
        if transform_id:
            transform = df_config.transforms_by(transform_id=transform_id)
        filename = DfReader.df_filename(df_config=df_config, df_id=df_id, transform=transform)
        result = self._df_dir_path(df_id=df_id, filename=filename)

        return result

    @staticmethod
    def df_filename(df_config: DfConfig, df_id: str, transform: Optional[DfTransformConfig] = None):
        """
        Forms a filename for the dataframe

        Parameters
        ----------
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        df_id: str
            Unique identifier of your dataframe.
        transform: optional DfTransformConfig
            Config representing a transform

        Returns
        -------
        str, a filename for the dataframe.
        """
        if transform:
            return f"{transform.transform_id}_{df_id}.{transform.df_format}"

        return f"{df_id}.{df_config.initial_df_format}"

    def _df_dir_path(self, df_id: str, filename: Optional[str] = None) -> str:
        """
        Forms an absolute file for a directory where you can find your dataframe or any other file.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        filename: optional str
            If you want to get a path for some particular file, provide its filename.

        Returns
        -------
        str, absolute path to the desired item
        """
        path = self._df_dir_path_for(dir_path=self._dir_path, df_id=df_id)

        if filename:
            path = os.path.join(path, filename)

        return path

    @staticmethod
    def _df_dir_path_for(dir_path: str, df_id: str) -> str:
        return os.path.join(dir_path, df_id)

    def _get_config(self, df_id: str) -> DfConfig:
        """
        Gets config object for the given dataframe.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.

        Returns
        -------
        DfConfig instance.
        """
        result = self._get_config_for(dir_path=self._df_dir_path(df_id=df_id), df_id=df_id)
        return result

    @staticmethod
    def _get_config_for(dir_path: str, df_id: str) -> DfConfig:
        """
        Gets config object for the given dataframe at the given path.

        Parameters
        ----------
        dir_path: str
            Absolute path to the dir when the dataframe is located.
        df_id: str
            Unique identifier of your dataframe.

        Returns
        -------

        """
        result = DfConfig(df_id=df_id, dir_path=dir_path)
        return result

    def _get_df_cache(self, df_format: str) -> DfCache:
        """
        Gives proper DfCache subclass for the given data format

        Parameters
        ----------
        df_format: str
            Format extension you want to use to save/load a dataframe.

        Returns
        -------
        DfCache instance.
        """
        df_cache = self._format_to_cache_map.get(df_format)
        if not df_cache:
            raise ValueError(f"Unknown df_format, df_cache was not provided: {df_format}")
        return df_cache



import pytest
import copy
from df_and_order.df_config import (
    DfConfig,
    DF_ID_KEY,
    CONFIG_FILENAME,
    DF_INITIAL_FORMAT_KEY,
    METADATA_KEY,
    TRANSFORMS_KEY,
)
from df_and_order.df_transform import DfTransformConfig
from df_and_order.df_transform_step import DfTransformStepConfig


@pytest.fixture()
def df_id():
    return "test_id"


@pytest.fixture()
def test_path():
    return "path/to/config/"


@pytest.fixture
def transformed_format():
    return "trans_format"


@pytest.fixture()
def transform(transformed_format):
    transform = DfTransformConfig(
        transform_id="transform_id",
        df_format=transformed_format,
        in_memory_steps=[DfTransformStepConfig(module_path="123", params={"a": 1, "b": "b"})],
        permanent_steps=[DfTransformStepConfig(module_path="456", params={"c": 2, "d": 1.2})],
    )
    return transform


def test_df_id_match(mocker, df_id, test_path):
    valid_df_id = df_id
    mocker.patch.object(
        DfConfig, DfConfig._read_config.__name__, lambda _, path: {DF_ID_KEY: valid_df_id},
    )
    DfConfig(df_id=df_id, dir_path=test_path)


def test_df_id_mismatch(mocker, df_id, test_path):
    wrong_df_id = "wrong_id"
    mocker.patch.object(
        DfConfig, DfConfig._read_config.__name__, lambda _, path: {DF_ID_KEY: wrong_df_id},
    )
    with pytest.raises(Exception):
        DfConfig(df_id=df_id, dir_path=test_path)


def test_config_exists(mocker, df_id, test_path):
    config_path = test_path + CONFIG_FILENAME
    mocker.patch.object(DfConfig, DfConfig._is_file_exists.__name__, lambda path: path == config_path)
    assert DfConfig.config_exists(dir_path=test_path)


def test_create_config_when_existing(mocker, df_id, test_path):
    mocker.patch.object(DfConfig, DfConfig.config_exists.__name__, lambda dir_path: True)
    save_mock = mocker.patch.object(DfConfig, DfConfig._save_at_path.__name__)

    with pytest.raises(Exception):
        DfConfig.create_config(dir_path=test_path, df_id=df_id, initial_df_format="")

    save_mock.assert_not_called()


@pytest.mark.parametrize("metadata", [{"meta": "shmeta"}, None], ids=["meta", "no_meta"])
@pytest.mark.parametrize("use_transform", [True, False], ids=["transform", "no_transform"])
def test_create_config(mocker, df_id, test_path, transform, metadata, use_transform):
    initial_df_format = "init_format"

    save_mock = mocker.patch.object(DfConfig, DfConfig._save_at_path.__name__)

    DfConfig.create_config(
        dir_path=test_path,
        df_id=df_id,
        initial_df_format=initial_df_format,
        metadata=metadata,
        transform=transform if use_transform else None,
    )

    config_dict = {
        DF_ID_KEY: df_id,
        DF_INITIAL_FORMAT_KEY: initial_df_format,
    }

    if metadata:
        config_dict[METADATA_KEY] = metadata

    if use_transform:
        transform_id, transform_dict = transform.to_dict()
        config_dict[TRANSFORMS_KEY] = {transform_id: transform_dict}

    save_mock.assert_called_with(config_dict=config_dict, config_path=test_path + CONFIG_FILENAME)


def test_properties(mocker, df_id, test_path):
    initial_df_format = "init_format"
    config_dict = {
        DF_ID_KEY: df_id,
        DF_INITIAL_FORMAT_KEY: initial_df_format,
    }
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)

    assert config.initial_df_format == initial_df_format
    assert config.df_id == df_id


def test_transform_by_no_transforms(mocker, df_id, test_path):
    config_dict = {
        DF_ID_KEY: df_id,
    }
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)

    with pytest.raises(Exception):
        config.transforms_by(transform_id="whatever")


def test_transform_by_no_transform(mocker, df_id, test_path):
    config_dict = {DF_ID_KEY: df_id, TRANSFORMS_KEY: {"some_transform_id": {}}}
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)

    with pytest.raises(Exception):
        config.transforms_by(transform_id="whatever")


def test_transform_by(mocker, df_id, test_path, transform):
    transform_id, transform_dict = transform.to_dict()

    config_dict = {DF_ID_KEY: df_id, TRANSFORMS_KEY: {transform_id: transform_dict}}
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)

    transform_result = config.transforms_by(transform_id=transform_id)
    _, transform_result_dict = transform_result.to_dict()

    assert transform_dict == transform_result_dict


def test_register_transform_already_cached(mocker, df_id, test_path, transform, transformed_format):
    initial_df_format = "init_format"
    transform_id, transform_dict = transform.to_dict()
    config_dict = {
        DF_ID_KEY: df_id,
        DF_INITIAL_FORMAT_KEY: initial_df_format,
        TRANSFORMS_KEY: {transform_id: transform_dict},
    }
    transfom_filename = f"{transform_id}_{df_id}"
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=config_dict)

    test_path_to_cached_file = test_path + transfom_filename + f".{transformed_format}"
    is_file_exists_mock = mocker.patch.object(DfConfig, DfConfig._is_file_exists.__name__)
    is_file_exists_mock.return_value = True

    with pytest.raises(Exception):
        config.register_transform(transform=transform, filename=transfom_filename)

    is_file_exists_mock.assert_called_with(path=test_path_to_cached_file)


def test_register_transform(mocker, df_id, test_path, transform, transformed_format):
    save_mock = mocker.patch.object(DfConfig, DfConfig._save_at_path.__name__)
    initial_df_format = "init_format"
    transform_id, transform_dict = transform.to_dict()
    config_dict = {
        DF_ID_KEY: df_id,
        DF_INITIAL_FORMAT_KEY: initial_df_format,
        TRANSFORMS_KEY: {transform_id: transform_dict},
    }
    config = _get_config(mocker=mocker, df_id=df_id, dir_path=test_path, config_dict=copy.deepcopy(config_dict),)

    updated_transform = DfTransformConfig(
        transform_id=transform_id,
        df_format=transformed_format,
        in_memory_steps=[DfTransformStepConfig(module_path="razdva", params={"asd": 123})],
    )

    config.register_transform(transform=updated_transform, filename="whatever")

    _, updated_transform_config = updated_transform.to_dict()

    config_dict[TRANSFORMS_KEY] = {transform_id: updated_transform_config}
    save_mock.assert_called_with(config_dict=config_dict, config_path=test_path + CONFIG_FILENAME)


def _get_config(mocker, df_id, dir_path, config_dict: dict):
    mocker.patch.object(DfConfig, DfConfig._read_config.__name__, lambda _, path: config_dict)
    config = DfConfig(df_id=df_id, dir_path=dir_path)
    return config



from typing import List, Optional, Tuple, Any, Dict

from df_and_order.df_transform_step import DfTransformStepConfig


TRANSFORM_ID_KEY = "transform_id"
TRANSFORM_DF_FORMAT_KEY = "df_format"
TRANSFORM_SOURCE_ID_KEY = "source_id"
TRANSFORM_SOURCE_IN_MEMORY_KEY = "source_in_memory"
TRANSFORM_IN_MEMORY_KEY = "in_memory"
TRANSFORM_PERMANENT_KEY = "permanent"


class DfTransformConfig:
    """

    Describes how a transformation should be performed.
    For any transformation one or many steps can be used.
    Those steps can be performed in memory only or their result
    can be serialized to the disk. Either in_memory or permanent or both steps
    should be provided in the init method.

    Parameters
    ----------
    transform_id: str
        Unique identifier of a transform.
    df_format: str
        What format to use to save permanent transforms
    source_id: optional str
        A transform may be based on another transform, source_id will be
        the desired transform_id you want to base on. If None - initial dataframe
        will be used by default.
    source_in_memory_steps: list of DfTransformStepConfig
        If source_id is not empty you may apply some transforms on a source_df first.
        It is useful when dealing with the initial dataframe.
    in_memory_steps: list of DfTransformStepConfig
        Those steps that are performed every time we read a dataframe from the disk.
    permanent_steps: list of DfTransformStepConfig
        Those steps result of which is persisted on the disk for future access.
    """

    def __init__(
        self,
        transform_id: str,
        df_format: str,
        source_id: Optional[str] = None,
        source_in_memory_steps: Optional[List[DfTransformStepConfig]] = None,
        in_memory_steps: Optional[List[DfTransformStepConfig]] = None,
        permanent_steps: Optional[List[DfTransformStepConfig]] = None,
    ):
        if in_memory_steps is None and permanent_steps is None:
            raise Exception("Provide at least one type of transformations")

        self._transform_id = transform_id
        self._df_format = df_format
        self._source_id = source_id
        self._source_in_memory_steps = source_in_memory_steps or []
        self._in_memory_steps = in_memory_steps or []
        self._permanent_steps = permanent_steps or []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DfTransformConfig):
            return False

        result = (
            self.transform_id == other.transform_id
            and self.df_format == other.df_format
            and self.source_id == other.source_id
            and self.source_in_memory_steps == other.source_in_memory_steps
            and self.in_memory_steps == other.in_memory_steps
            and self.permanent_steps == other.permanent_steps
        )

        return result

    @property
    def transform_id(self) -> str:
        return self._transform_id

    @property
    def df_format(self) -> str:
        return self._df_format

    @property
    def source_id(self) -> Optional[str]:
        return self._source_id

    @property
    def source_in_memory_steps(self) -> List[DfTransformStepConfig]:
        return self._source_in_memory_steps

    @property
    def in_memory_steps(self) -> List[DfTransformStepConfig]:
        return self._in_memory_steps

    @property
    def permanent_steps(self) -> List[DfTransformStepConfig]:
        return self._permanent_steps

    def to_dict(self) -> Tuple[str, dict]:
        """
        Converts an instance to serializable dictionary

        Returns
        -------
        tuple of str and dict, transform_id and transform's dictionary representation.
        """
        result: Dict[str, Any] = {TRANSFORM_DF_FORMAT_KEY: self._df_format}

        if self._source_id:
            result[TRANSFORM_SOURCE_ID_KEY] = self._source_id

        if self._source_in_memory_steps:
            source_in_memory_dicts = [config.to_dict() for config in self._source_in_memory_steps]
            result[TRANSFORM_SOURCE_IN_MEMORY_KEY] = source_in_memory_dicts

        if self._in_memory_steps:
            in_memory_dicts = [config.to_dict() for config in self._in_memory_steps]
            result[TRANSFORM_IN_MEMORY_KEY] = in_memory_dicts

        if self._permanent_steps:
            permanent_dicts = [config.to_dict() for config in self._permanent_steps]
            result[TRANSFORM_PERMANENT_KEY] = permanent_dicts

        return self.transform_id, result

    @staticmethod
    def from_dict(transform_id: str, transform_dict: dict):
        """
        Builds DfTransformConfig instance out of serialized dictionary.

        Parameters
        ----------
        transform_id: optional str
            Unique identifier of a transform.
        transform_dict:
            Dictionary representation of DfTransformConfig

        Returns
        -------
        DfTransformConfig instance.
        """
        step_configs_getter = DfTransformConfig._transform_step_configs_from
        df_format = transform_dict.get(TRANSFORM_DF_FORMAT_KEY)
        if df_format is None:
            raise Exception(
                "Provide a file format you want your transform " "to be saved in using df_format key in the config"
            )

        source_id = transform_dict.get(TRANSFORM_SOURCE_ID_KEY)
        source_in_memory_transforms = step_configs_getter(
            transform_dict=transform_dict, key=TRANSFORM_SOURCE_IN_MEMORY_KEY
        )
        in_memory_transforms = step_configs_getter(transform_dict=transform_dict, key=TRANSFORM_IN_MEMORY_KEY)
        permanent_transforms = step_configs_getter(transform_dict=transform_dict, key=TRANSFORM_PERMANENT_KEY)

        result = DfTransformConfig(
            transform_id=transform_id,
            df_format=df_format,
            source_id=source_id,
            source_in_memory_steps=source_in_memory_transforms,
            in_memory_steps=in_memory_transforms,
            permanent_steps=permanent_transforms,
        )
        return result

    @staticmethod
    def _transform_step_configs_from(transform_dict: dict, key: str) -> Optional[List[DfTransformStepConfig]]:
        """
        When deserializing we need to transform steps into objects as well.

        Parameters
        ----------
        transform_dict:
            Dictionary representation of DfTransformConfig
        key: str
            Key for transform steps.

        Returns
        -------
        Optional list of DfTransformStepConfig objects.
        """
        step_dicts = transform_dict.get(key)
        steps = None
        if step_dicts:
            steps = [DfTransformStepConfig.from_dict(step_dict=step_dict) for step_dict in step_dicts]

        return steps



import pandas as pd
from typing import List


from df_and_order.df_transform_step import DfTransformStep


class TestDatesTransformStep(DfTransformStep):
    def __init__(self, dates_cols: List[str]):
        super().__init__()

        self._dates_cols = dates_cols

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._dates_cols:
            df[col] = pd.to_datetime(df[col])

        return df



import pandas as pd
import pytest

from df_and_order.df_cache import DfCache


@pytest.fixture()
def save_args():
    return [1, 2, 3]


@pytest.fixture()
def save_kwargs():
    return {"a": 1, "b": 2}


@pytest.fixture()
def load_args():
    return [4, 5, 6]


@pytest.fixture()
def load_kwargs():
    return {"c": 3, "d": 4}


class TestDfCache(DfCache):
    def _save(self, df: pd.DataFrame, path: str, *args, **kwargs):
        pass

    def _load(self, path: str, *args, **kwargs) -> pd.DataFrame:
        pass


@pytest.fixture()
def instance(save_args, save_kwargs, load_args, load_kwargs):
    result = TestDfCache(save_args=save_args, save_kwargs=save_kwargs, load_args=load_args, load_kwargs=load_kwargs,)
    return result


def test_save(mocker, instance, save_args, save_kwargs):
    save_mock = mocker.patch.object(instance, "_save")
    df = mocker.Mock()
    path = "/some/path"
    instance.save(df=df, path=path)

    save_mock.assert_called_with(df=df, path=path, *save_args, **save_kwargs)


def test_load(mocker, instance, load_args, load_kwargs):
    load_mock = mocker.patch.object(instance, "_load")
    path = "/some/path"
    instance.load(path=path)

    load_mock.assert_called_with(path=path, *load_args, **load_kwargs)
